@@ -10,22 +10,23 @@ import sys
 import textwrap
 import types
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, AnyStr, Callable, ForwardRef, NewType, TypeVar, get_type_hints
+from typing import TYPE_CHECKING, Any, AnyStr, ForwardRef, NewType, TypeVar, get_type_hints
 
 from docutils import nodes
-from docutils.frontend import OptionParser
+from docutils.frontend import get_default_settings
 from sphinx.ext.autodoc.mock import mock
 from sphinx.parsers import RSTParser
 from sphinx.util import logging, rst
 from sphinx.util.inspect import TypeAliasForwardRef, TypeAliasNamespace, stringify_signature
 from sphinx.util.inspect import signature as sphinx_signature
 
-from .parser import parse
+from ._parser import parse
 from .patches import install_patches
 from .version import __version__
 
 if TYPE_CHECKING:
     from ast import FunctionDef, Module, stmt
+    from collections.abc import Callable
 
     from docutils.nodes import Node
     from docutils.parsers.rst import states
@@ -63,7 +64,7 @@ def get_annotation_module(annotation: Any) -> str:
         return "builtins"
     if _get_types_type(annotation) is not None:
         return "types"
-    is_new_type = sys.version_info >= (3, 10) and isinstance(annotation, NewType)
+    is_new_type = isinstance(annotation, NewType)
     if (
         is_new_type
         or isinstance(annotation, TypeVar)
@@ -79,8 +80,6 @@ def get_annotation_module(annotation: Any) -> str:
 
 
 def _is_newtype(annotation: Any) -> bool:
-    if sys.version_info < (3, 10):
-        return inspect.isfunction(annotation) and hasattr(annotation, "__supertype__")
     return isinstance(annotation, NewType)
 
 
@@ -158,7 +157,7 @@ def get_annotation_args(annotation: Any, module: str, class_name: str) -> tuple[
 
 
 def format_internal_tuple(t: tuple[Any, ...], config: Config) -> str:
-    # An annotation can be a tuple, e.g., for nptyping:
+    # An annotation can be a tuple, e.g., for numpy.typing:
     # In this case, format_annotation receives:
     # This solution should hopefully be general for *any* type that allows tuples in annotations
     fmt = [format_annotation(a, config) for a in t]
@@ -169,6 +168,18 @@ def format_internal_tuple(t: tuple[Any, ...], config: Config) -> str:
     return f"({', '.join(fmt)})"
 
 
+def fixup_module_name(config: Config, module: str) -> str:
+    if getattr(config, "typehints_fixup_module_name", None):
+        module = config.typehints_fixup_module_name(module)
+
+    if module == "typing_extensions":
+        module = "typing"
+
+    if module == "_io":
+        module = "io"
+    return module
+
+
 def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PLR0911, PLR0912, PLR0915, PLR0914
     """
     Format the annotation.
@@ -204,13 +215,7 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     except ValueError:
         return str(annotation).strip("'")
 
-    # Redirect all typing_extensions types to the stdlib typing module
-    if module == "typing_extensions":
-        module = "typing"
-
-    if module == "_io":
-        module = "io"
-
+    module = fixup_module_name(config, module)
     full_name = f"{module}.{class_name}" if module != "builtins" else class_name
     fully_qualified: bool = getattr(config, "typehints_fully_qualified", False)
     prefix = "" if fully_qualified or full_name == class_name else "~"
@@ -228,7 +233,7 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     # Some types require special handling
     if full_name == "typing.NewType":
         args_format = f"\\(``{annotation.__name__}``, {{}})"
-        role = "class" if sys.version_info >= (3, 10) else "func"
+        role = "class"
     elif full_name in {"typing.TypeVar", "typing.ParamSpec"}:
         params = {k: getattr(annotation, f"__{k}__") for k in ("bound", "covariant", "contravariant")}
         params = {k: v for k, v in params.items() if v}
@@ -349,9 +354,24 @@ def process_signature(  # noqa: C901, PLR0913, PLR0917
 
     obj = inspect.unwrap(obj)
     sph_signature = sphinx_signature(obj, type_aliases=app.config["autodoc_type_aliases"])
+    typehints_formatter: Callable[..., str | None] | None = getattr(app.config, "typehints_formatter", None)
+
+    def _get_formatted_annotation(annotation: TypeVar) -> TypeVar:
+        if typehints_formatter is None:
+            return annotation
+        formatted_name = typehints_formatter(annotation)
+        return annotation if not isinstance(formatted_name, str) else TypeVar(formatted_name)
+
+    if app.config.typehints_use_signature_return:
+        sph_signature = sph_signature.replace(
+            return_annotation=_get_formatted_annotation(sph_signature.return_annotation)
+        )
 
     if app.config.typehints_use_signature:
-        parameters = list(sph_signature.parameters.values())
+        parameters = [
+            param.replace(annotation=_get_formatted_annotation(param.annotation))
+            for param in sph_signature.parameters.values()
+        ]
     else:
         parameters = [param.replace(annotation=inspect.Parameter.empty) for param in sph_signature.parameters.values()]
 
@@ -373,7 +393,7 @@ def process_signature(  # noqa: C901, PLR0913, PLR0917
                 # when method starts with double underscore Python applies mangling -> prepend the class name
                 method_name = f"_{obj.__qualname__.split('.')[-2]}{method_name}"
             method_object = outer.__dict__[method_name] if outer else obj
-            if not isinstance(method_object, (classmethod, staticmethod)):
+            if not isinstance(method_object, classmethod | staticmethod):
                 start = 1
 
     sph_signature = sph_signature.replace(parameters=parameters[start:])
@@ -397,14 +417,13 @@ def _is_dataclass(name: str, what: str, qualname: str) -> bool:
 
 
 def _future_annotations_imported(obj: Any) -> bool:
-    _annotations = getattr(inspect.getmodule(obj), "annotations", None)
-    if _annotations is None:
+    annotations_ = getattr(inspect.getmodule(obj), "annotations", None)
+    if annotations_ is None:
         return False
 
     # Make sure that annotations is imported from __future__ - defined in cpython/Lib/__future__.py
     # annotations become strings at runtime
-    future_annotations = 0x100000 if sys.version_info[0:2] == (3, 7) else 0x1000000
-    return bool(_annotations.compiler_flag == future_annotations)
+    return bool(annotations_.compiler_flag == 0x1000000)  # pragma: no cover # noqa: PLR2004
 
 
 def get_all_type_hints(
@@ -653,7 +672,7 @@ def process_docstring(  # noqa: PLR0913, PLR0917
 
     localns = TypeAliasNamespace(app.config["autodoc_type_aliases"])
     type_hints = get_all_type_hints(app.config.autodoc_mock_imports, obj, name, localns)
-    app.config._annotation_globals = getattr(obj, "__globals__", {})  # type: ignore[attr-defined]  # noqa: SLF001
+    app.config._annotation_globals = getattr(obj, "__globals__", {})  # noqa: SLF001
     try:
         _inject_types_to_docstring(type_hints, signature, original_obj, app, what, name, lines)
     finally:
@@ -829,7 +848,7 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
 
     # 3. Insert after the parameters.
     # To find the parameters, parse as a docutils tree.
-    settings = OptionParser(components=(RSTParser,)).get_default_values()
+    settings = get_default_settings(RSTParser)
     settings.env = app.env
     doc = parse("\n".join(lines), settings)
 
@@ -849,7 +868,7 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
         # at end. (I don't know of any input where this happens.)
         next_sibling = child.next_node(descend=False, siblings=True)
         line_no = node_line_no(next_sibling) if next_sibling else None
-        at = line_no - 2 if line_no else len(lines)
+        at = max(line_no - 2, 0) if line_no else len(lines)
         return InsertIndexInfo(insert_index=at, found_param=True)
 
     # 4. Insert before examples
@@ -857,7 +876,7 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
         if tag_name(child) in {"literal_block", "paragraph", "field_list"}:
             continue
         line_no = node_line_no(child)
-        at = line_no - 2 if line_no else len(lines)
+        at = max(line_no - 2, 0) if line_no else len(lines)
         return InsertIndexInfo(insert_index=at, found_directive=True)
 
     # 5. Otherwise, insert at end
@@ -967,6 +986,7 @@ def setup(app: Sphinx) -> dict[str, bool]:
     app.add_config_value("typehints_formatter", None, "env")
     app.add_config_value("typehints_use_signature", False, "env")  # noqa: FBT003
     app.add_config_value("typehints_use_signature_return", False, "env")  # noqa: FBT003
+    app.add_config_value("typehints_fixup_module_name", None, "env")
     app.add_role("sphinx_autodoc_typehints_type", sphinx_autodoc_typehints_type_role)
     app.connect("env-before-read-docs", validate_config)  # config may be changed after “config-inited” event
     app.connect("autodoc-process-signature", process_signature)
